"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.RedisCollaborationSessionManager = void 0;
const defaultRedisCollaborationSessionManagerOptions = {
    maxScanBatchSize: 800,
};
/**
 * Manages the set of collaboration sessions in a Redis hashmap.
 * @internal
 */
class RedisCollaborationSessionManager {
    constructor(redisClientConnectionManager, parameters, options) {
        this.redisClientConnectionManager = redisClientConnectionManager;
        /**
         * Redis hashmap key.
         */
        this.prefix = "collaboration-session";
        this.options = { ...defaultRedisCollaborationSessionManagerOptions, ...options };
        if (parameters?.prefix) {
            this.prefix = parameters.prefix;
        }
        redisClientConnectionManager.addErrorHandler(undefined, // lumber properties
        "Collaboration Session Manager Redis Error");
    }
    async addOrUpdateSession(session) {
        const key = this.getFieldKey(session);
        await this.redisClientConnectionManager
            .getRedisClient()
            .hset(this.prefix, key, JSON.stringify(this.getShortSession(session)));
    }
    async removeSession(session) {
        const key = this.getFieldKey(session);
        await this.redisClientConnectionManager.getRedisClient().hdel(this.prefix, key);
    }
    async getSession(session) {
        const key = this.getFieldKey(session);
        const sessionJson = await this.redisClientConnectionManager
            .getRedisClient()
            .hget(this.prefix, key);
        if (sessionJson === null) {
            return undefined;
        }
        return this.getFullSession(key, JSON.parse(sessionJson));
    }
    async getAllSessions() {
        // Use HSCAN to iterate over te key:value pairs of the hashmap
        // in batches to get all sessions with minimal impact on Redis.
        const sessionJsonScanStream = this.redisClientConnectionManager
            .getRedisClient()
            .hscanStream(this.prefix, { count: this.options.maxScanBatchSize });
        return new Promise((resolve, reject) => {
            const sessions = [];
            sessionJsonScanStream.on("data", (result) => {
                if (!result) {
                    // When redis scan is done, it pushes null to the stream.
                    // This should only trigger the "end" event, but we should check for it to be safe.
                    return;
                }
                if (!Array.isArray(result) || result.length % 2 !== 0) {
                    throw new Error("Invalid scan result");
                }
                // The scan stream emits an array of alternating field keys and values.
                // For example, a hashmap like { key1: value1, key2: value2 } would emit ["key1", "value1", "key2", "value2"].
                for (let i = 0; i < result.length; i += 2) {
                    const fieldKey = result[i];
                    const sessionJson = result[i + 1];
                    sessions.push(this.getFullSession(fieldKey, JSON.parse(sessionJson)));
                }
            });
            sessionJsonScanStream.on("end", () => {
                resolve(sessions);
            });
            sessionJsonScanStream.on("error", (error) => {
                reject(error);
            });
        });
    }
    getShortSession(session) {
        return {
            fjt: session.firstClientJoinTime,
            llt: session.lastClientLeaveTime,
            tp: {
                hwc: session.telemetryProperties.hadWriteClient,
                tlj: session.telemetryProperties.totalClientsJoined,
                mcc: session.telemetryProperties.maxConcurrentClients,
            },
        };
    }
    getFullSession(fieldKey, shortSession) {
        return {
            ...this.getTenantIdDocumentIdFromFieldKey(fieldKey),
            firstClientJoinTime: shortSession.fjt,
            lastClientLeaveTime: shortSession.llt,
            telemetryProperties: {
                hadWriteClient: shortSession.tp.hwc,
                totalClientsJoined: shortSession.tp.tlj,
                maxConcurrentClients: shortSession.tp.mcc,
            },
        };
    }
    getFieldKey(session) {
        return `${session.tenantId}:${session.documentId}`;
    }
    getTenantIdDocumentIdFromFieldKey(fieldKey) {
        const [tenantId, documentId] = fieldKey.split(":");
        return { tenantId, documentId };
    }
}
exports.RedisCollaborationSessionManager = RedisCollaborationSessionManager;
//# sourceMappingURL=redisSessionManager.js.map